--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      Class MediaPlayer - win version based on DirectShow xtra
-- @author    Valentin Schmidt
-- @version   0.2
--****************************************************************************

global $

-- wrapped xtra instance
property _dx

-- callback
property _cbMessageHandler
property _cbMessageTarget

-- player states
property _fullscreen
property _pausedAtStart -- if files are paused at start (after opening) or not
property _loop
property _volume
property _balance
-- win only stuff
property _keep_ratio
property _registergraph

property _hasVideo
property _isMuted

----------------------------------------
-- Creates a new player instance.
-- @constructor
-- @param {propList|VOID} [settings=VOID] - player start settings; optional, if not specified, defaults are used
-- @param {symbol|VOID} [cbMessageHandler]
-- @param {object} [cbMessageTarget=_movie]
----------------------------------------
on new (me, settings, cbMessageHandler, cbMessageTarget)

  -- DirectShow xtra wrapper, enhanced with "detached" callbacks and named contsants
  me._dx = $.import("directshow", settings)

  me._hasVideo = FALSE
  me._isMuted = FALSE

  -- optional player states on start
  if voidP(settings) then settings = [:]

  me._fullscreen = (settings["fullscreen"]=1)
  me._pausedAtStart = (settings["pausedAtStart"]=1)
  me._loop = (settings["loop"]=1)

  if voidP(settings["volume"]) then me._volume = 0.8
  else me._volume = settings["volume"]

  if voidP(settings["balance"]) then me._balance = 0.0
  else me._balance = settings["balance"]

  me._keep_ratio = (settings["keep_ratio"]=1) -- win only
  me._registergraph = (settings["registergraph"]=1) -- win only

  -- event callback
  if not voidP(cbMessageHandler) then
    if voidP(cbMessageTarget) then cbMessageTarget = _movie
    me._cbMessageHandler = cbMessageHandler
    me._cbMessageTarget = cbMessageTarget

    -- setup callbacks ("detached")
    me._dx.setCallback(#_directShowEvent, me)
  end if

  return me
end

----------------------------------------
-- Cleans up on exit (not really needed).
----------------------------------------
on mExit (me)
  me.mCloseFile()
  me._dx = VOID
end

----------------------------------------
-- Tries to render a media file with enclosed LAV filters.
-- Optionally can also load subtitles into a video.
-- @param {string} mediaFile
-- @param {propList} [settings] - optional, allows to overwrite current player states
-- @param {string|TRUE} [subtitles] - optional, either path to SRT file or TRUE for auto-detection of subtitle files
-- @return {TRUE|list} either returns TRUE or list with error infos
----------------------------------------
on mOpenFile (me, mediaFile, settings, subtitles)
  --me.mCloseFile() -- close previous

  -- overwrite current player states?
  if ilk(settings)=#propList then
    if not voidP(settings["balance"]) then me._balance = settings["balance"]
    if not voidP(settings["fullscreen"]) then me._fullscreen = settings["fullscreen"]
    if not voidP(settings["keep_ratio"]) then me._keep_ratio = settings["keep_ratio"]
    if not voidP(settings["loop"]) then me._loop = settings["loop"]
    if not voidP(settings["pausedAtStart"]) then me._pausedAtStart = settings["pausedAtStart"]
    if not voidP(settings["registergraph"]) then me._registergraph = settings["registergraph"]
    --if not voidP(settings["volume"]) then me._volume = settings["volume"]
  end if

  props = [:]

  props["balance"] = 10000*me._balance/5
  props["loop"] = me._loop
  props["paused"] = me._pausedAtStart
  props["registergraph"] = me._registergraph

  if me._isMuted then
    props["volume"] = -10000
  else
    props["volume"] = integer(1442.6950 * log(me._volume))
  end if

  props["mouseevents"] = 1

  -- create filter graph
  props["filterlist"] = [:]

  -- add LAV filters
  props["filterlist"]["filters"] = []
  props["filterlist"]["filters"].add(["guid":me._dx.CLSID.LAVSplitterSource,\
    "filterpath":$.PATH&"filters\LAVSplitter.ax",\
    "source":mediaFile])
  props["filterlist"]["filters"].add(["guid":me._dx.CLSID.LAVVideoDecoder,\
    "filterpath":$.PATH&"filters\LAVVideo.ax"])
  props["filterlist"]["filters"].add(["guid":me._dx.CLSID.LAVAudioDecoder,\
    "filterpath":$.PATH&"filters\LAVAudio.ax"])

  -- show subtitles via VSFilter/DirectVobSub/DirectVobSubXy
  if subtitles=TRUE then -- autoloading subtitles via filename
    props["filterlist"]["filters"].add(["guid":me._dx.CLSID.VSFilter_autoload,\
      "filterpath":$.PATH&"filters\VSFilter.dll"])
  else if stringP(subtitles) then -- subtitle file specified by path
    props["filterlist"]["filters"].add(["guid":me._dx.CLSID.VSFilter,\
      "filterpath":$.PATH&"filters\VSFilter.dll"])
  end if

  -- list that allows to test 3 different variations
  render = [:]
  render[#video_audio] = ["connections":[[1,2],[1,3]], "render":[2,3]]
  render[#video_only] =  ["connections":[[1,2]], "render":[2]]
  render[#audio_only] =  ["connections":[[1,3]], "render":[3]]

  repeat with i = 1 to render.count --r in render
    r = render[i]
    props["filterlist"]["connections"] = r["connections"]
    props["filterlist"]["render"] = r["render"]
    hr = me._dx.ds_open(props)
    if hr=0 then exit repeat -- success!
  end repeat

  if hr<0 then
    return me._dx.ds_error_info(hr)
  else

    me._hasVideo = (i=1 or i=2)

    -- load subtitleFile, if specified
    if stringP(subtitles) then
      me._dx.ds_subtitles_load(subtitles)
    end if

    -- force video screen update in paused state
    if me._pausedAtStart then me._dx.ds_step_forward(0)

    --me._sendMessage(#opened)

    -- test
    if me._fullscreen then me._dx.ds_set_param(["fullscreen":TRUE])

    return TRUE
  end if
end

----------------------------------------
-- Closes the currently opened media file.
----------------------------------------
on mCloseFile (me)
  me._dx.ds_close()
end

----------------------------------------
-- Changes the rect of the current video.
-- @param {rect} videoRect
----------------------------------------
on mSetVideoRect (me, videoRect)
  videoRect = videoRect.inflate(1,1)

  -- keep aspect ratio
  if me._keep_ratio then
    info = me._dx.ds_get_media_info()
    if not voidP(info) then
      if info["width"]>0 AND info["height"]>0 then
        ratio = float(info["width"])/info["height"]
        videoRect = me._findMaxRect(videoRect, ratio)
      end if
    end if
  end if

  me._dx.ds_set_param(["rect": videoRect])
end

----------------------------------------
-- Zooms a video, relative to its native size.
-- Returns either the new rect for the zoomed video, or VOID in case of failure.
-- @param {float} zoomFactor - the zoom factor, either 0.5, 1.0 or 2.0
-- @return {rect|VOID}
----------------------------------------
on mZoom (me, zoomFactor)
  stat = me._dx.ds_get_status()
  if stat=0 then return -- no media loaded
  me._dx.ds_set_param(["zoom": zoomFactor])
  me._videoRect = me._dx.ds_get_param("rect")
  return me._videoRect
end

----------------------------------------
-- Toggles fullscreen mode.
----------------------------------------
on mToggleFullscreen (me)
  if not me._hasVideo then return
  me._fullscreen = not me._fullscreen
  props = ["fullscreen":me._fullscreen]
  me._dx.ds_set_param(props)
end

----------------------------------------
-- Toggles visibility of currently loaded video.
----------------------------------------
on mToggleVisible (me)
  vis = not me._dx.ds_get_param("visible")
  me._dx.ds_set_param(["visible": vis])
  return vis
end

----------------------------------------
-- Starts/resumes playback of currently loaded media file.
----------------------------------------
on mPlay (me)
  me._dx.ds_play()
end

----------------------------------------
-- Pauses playback of currently loaded media file.
----------------------------------------
on mPause (me)
  me._dx.ds_pause()
end

-- new
on mStop (me)
  me._dx.ds_stop()
  me.mSetCurrentTime(0)
end

----------------------------------------
-- Steps one frame backwards (if used filters support this).
----------------------------------------
on mStepBackwards (me)
  me._dx.ds_step_backwards()
end

----------------------------------------
-- Steps one frame forward (if used filters support this).
----------------------------------------
on mStepForward (me)
  me._dx.ds_step_forward()
end

----------------------------------------
-- Changes the current time (in ms) of currently loaded media file.
-- @param {integer} timeMS
----------------------------------------
on mSetCurrentTime (me, timeMS)
  me._dx.ds_set_param(["time":timeMS])
end

----------------------------------------
-- Returns the current time (in ms) of currently loaded media file.
-- @return {integer}
----------------------------------------
on mGetCurrentTime (me)
  return me._dx.ds_get_param("time")
end

----------------------------------------
-- Changes the play rate of currently loaded media file (if used filters support this).
-- @param {float} rate
----------------------------------------
on mSetRate (me, rate)
  me._dx.ds_set_param(["rate": rate])
end

----------------------------------------
-- Toggles loop mode, i.e. if a media file automatically restarts after reaching its end.
----------------------------------------
on mToggleLoop (me)
  me._loop = not me._loop
  me._dx.ds_set_param(["loop": me._loop])
end

----------------------------------------
--
----------------------------------------
on mSetMuted (me, flag)
  me._isMuted = flag
  if flag then
    props = ["volume": -10000] --> -10000..0
    me._dx.ds_set_param(props)
  else
    me.mSetVolume(me._volume)
  end if
end

----------------------------------------
-- Sets the playback volume of currently loaded media file.
-- @param {float} vol - in range 0.0..1.0
----------------------------------------
on mSetVolume (me, vol)
  me._volume = vol

  if me._isMuted then return

  -- 0..1 --> -10000..0

  --if vol=0 then vol=-10000
  --else vol = integer( (vol-1)*3000 )

  if vol=0 then vol=-10000
  else vol = integer(1442.6950 * log(vol))

  props = ["volume": vol] --> -10000..0
  me._dx.ds_set_param(props)
end

----------------------------------------
-- Sets the audio balance (pan) of currently loaded media file.
-- @param {float} balance - in range -1.0..1.0
----------------------------------------
on mSetBalance (me, balance)
  me._balance = balance

  props = ["balance": integer(balance*10000)] -- -10,000..10,000.
  me._dx.ds_set_param(props)
end

----------------------------------------
-- Returns the duration (in ms) of the currently loaded media file.
-- @return {integer}
----------------------------------------
on mGetDuration (me)
  info = me._dx.ds_get_media_info()
  return info["duration"]
end

----------------------------------------
-- Returns the current playback status as symbol.
-- @return {symbol} either #idle, #paused, #running or #stopped
----------------------------------------
on mGetStatus (me)
  stat = me._dx.ds_get_status()
  return [#idle, #stopped, #paused, #running][stat+1]
end

----------------------------------------
-- Returns a propList (or VOID) with infos about the currently loaded media file.
-- The returned propList has the following properties:
--   "duration" (integer ms)
--   "width" (integer)
--   "height" (integer)
--   "fps" (float)
--   "tracks" (list, containing "vide" for video and "soun" for audio tracks)
-- @return {propList|VOID}
----------------------------------------
on mGetMediaInfos (me)
  infos = me._dx.ds_get_media_info()
  if voidP(infos) then return VOID

  -- new
  infos["hasVideo"] = me._hasVideo

--  infos["tracks"] = []
--  tc = me._dx.ds_get_stream_count(me._dx.CLSID.LAVSplitterSource)-1
--  repeat with i = 0 to tc
--    trackInfo = me._dx.ds_get_stream_info(me._dx.CLSID.LAVSplitterSource, i)
--    if trackInfo["majortype"]=me._dx.MEDIATYPE.Video then
--      infos["tracks"].add("vide")
--    else if trackInfo["majortype"]=me._dx.MEDIATYPE.Audio then
--      infos["tracks"].add("soun")
--    end if
--  end repeat

  return infos
end

--**************************************
-- PRIVATE
--**************************************

----------------------------------------
-- @callback
----------------------------------------
on _directShowEvent (me, evt, param1, param2)

  evtName = me._dx.EVENT.getOne(evt)
  --put evtName

  case evtName of

  #EC_FULLSCREEN_LOST:
    me._fullscreen = FALSE
    me._sendMessage(#fullscreenLost)

  #EC_COMPLETE:
    me._sendMessage(#complete) -- stopped | complete ?

  end case
end

----------------------------------------
--
----------------------------------------
on _sendMessage (me, stat)
  if not voidP(me._cbMessageHandler) then
    call(me._cbMessageHandler, me._cbMessageTarget, stat, me)
  end if
end

----------------------------------------
--
----------------------------------------
on _findMaxRect (me, tMaxRect, tRatio)
  if tMaxRect.width=0 OR tMaxRect.height=0 OR tRatio=0 then
    return tMaxRect
  end if
  tMaxRectRatio = float(tMaxRect.width)/tMaxRect.height
  if tMaxRectRatio>=tRatio then
    h = tMaxRect.height
    w = h * tRatio
    x = (tMaxRect.width-w)/2
    y = 0
  else
    w = tMaxRect.width
    h = w / tRatio
    x = 0
    y = (tMaxRect.height-h)/2
  end if
  return rect(x, y, x+w, y+h)
end
